constants = {
    "NoOp": ["int", 0],
    "OptIn": ["int", 1],
    "CloseOut": ["int", 2],
    "ClearState": ["int", 3],
    "UpdateApplication": ["int", 4],
    "DeleteApplication": ["int", 5],
    "Pay": ["int", 1],
    "Acfg": ["int", 3],
    "Axfer": ["int", 4],
    "Afrz": ["int", 5],
    "Appl": ["int", 6],
}
